/* $Header */
/*
 * Copyright (C) 1995 Premia Corporation	
 */
 
#include <windows.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "exports.h"
#include "cwstart.h"

#define RETAB	(1)			// do conversion for this buffer/file
#define NORETAB	(2)			// don't convert this buffer/file


long DLL Tab_new_handler   ( long,  long  );
long DLL Tab_save_handler  ( long,  long  );
long DLL Tab_saved_handler ( long,  long  );
BOOL DLL TabSetConvert     ( int    );   
BOOL DLL TabInstallHandler ( void   );   
BOOL DLL TabRemoveHandler  ( void   );   
void DLL DeTabBuffer       ( LPSTR  );   
void DLL EnTabBuffer       ( LPSTR  );   
void DLL ReTabBuffer       ( LPSTR, LPSTR, BOOL );
BOOL DLL TabConfirm        ( BOOL   );   
int  DLL TabSpacesOnly     ( int    );   


static LPVOID tabHashPtr    = NULL;
static HEVENT eva[3]        = {0,}; 
static BOOL   tabConfirm    = TRUE; 		//   prompt before any conversions
//   How to save whitespace: 1 = tabs+spaces; -1 = spacesonly; 0 = use buf flags
static int    tabSpacesOnly = 0;		
static LPSTR  dosTabStr     = "9 17";

static LPINT tabAToI   ( LPSTR tabStr );   
static LPSTR DeTabLine ( LPSTR iLine, LPINT tabArr     );   
static LPSTR EnTabLine ( LPSTR iLine, LPINT tabArr     );   
static LPSTR ReTabLine ( LPSTR iLine, LPINT oldTabArr, 
						 LPINT newTabArr, BOOL useSpaces );
static int   nextTab   ( int   col,   LPINT tabArr     );   


/*
 *  The high-level functions use tab strings.  The
 *  lower-level functions use tab arrays for speed.
 *
 *
 */


LIBMAIN

void DLL 
_init()
{
	LibExport( "long Tab_new_handler   long  long  "  );
	LibExport( "long Tab_save_handler  long  long  "  );
	LibExport( "long Tab_saved_handler long  long  "  );
	LibExport( "BOOL TabSetConvert     int   "     );
	LibExport( "BOOL TabInstallHandler "     );   
	LibExport( "BOOL TabRemoveHandler  "     );   
	LibExport( "void DeTabBuffer       LPSTR "     );
	LibExport( "void EnTabBuffer       LPSTR "     );
	LibExport( "void ReTabBuffer       LPSTR LPSTR BOOL "  );
	LibExport( "void EnTabThisLine     "     );   
	LibExport( "void DeTabThisLine     "     );   
	LibExport( "BOOL TabConfirm        BOOL  "     );
	LibExport( "BOOL TabSpacesOnly     BOOL  "     );
	LibExport( "int  NextTab           int   LPSTR "  );

	tabHashPtr = HashCreateTable(41, sizeof(int), HASH_IGN_CASE);

} 

int DLL
WEP(int nParam)
{
	TabRemoveHandler();

	return(0);
} 

BOOL DLL
TabConfirm(BOOL confirm)
{
	return(tabConfirm = !!confirm);
}

int DLL
TabSpacesOnly(int spacesOnly)
{
	if (spacesOnly > 0)
		tabSpacesOnly = 1;
	else if (spacesOnly < 0)
		tabSpacesOnly = -1;
	else	
		tabSpacesOnly = 0;
	return(tabSpacesOnly);
}

void DLL
EntabThisLine(void)
{
	LPINT tabArr;
	LPSTR thisLine;
	LPSTR thatLine;
	
	MarkSavePos();
	MovHome();
	thisLine = BufReadStr(0x7FFFFFFF);
	tabArr = tabAToI(BufQTabStr());
	thatLine = EnTabLine(thisLine, tabArr);
	if (thatLine && thatLine != thisLine)
	{
		BufDelToEOL();
		BufInsertStr(thatLine);
		StrFree(thatLine);
	}
	else
	{
		MsgPrintf(2, "Line is already entabbed");
	}
	StrFree(thisLine);
	MemFree(tabArr);
	MarkRestorePos();
}

void DLL
DetabThisLine(void)
{
	LPINT tabArr;
	LPSTR thisLine;
	LPSTR thatLine;
	
	MarkSavePos();
	MovHome();
	thisLine = BufReadStr(0x7FFFFFFF);
	tabArr = tabAToI(BufQTabStr());
	thatLine = DeTabLine(thisLine, tabArr);
	if (thatLine && thatLine != thisLine)
	{
		BufDelToEOL();
		BufInsertStr(thatLine);
		StrFree(thatLine);
	}
	else
	{
		MsgPrintf(2, "Line is already detabbed");
	}
	StrFree(thisLine);
	MemFree(tabArr);
	MarkRestorePos();
}

void DLL
ReTabBuffer(LPSTR oldTabStr, LPSTR newTabStr, BOOL useSpaces)
{
	LPSTR newBuf;   
	LPSTR oldBuf;   
	long  curLine    = 1;
	long  totalLines = 0;
	LPINT oldTabArr, newTabArr;
	
	if (ArgCount() >= 0)
	{
		if (ArgCount() < 2 || !newTabStr)
			newTabStr = BufQTabStr();
		if (ArgCount() < 1 || !oldTabStr)
			oldTabStr = dosTabStr;
	}	
	else
	{
		if (!newTabStr)
			newTabStr = BufQTabStr();
		if (!oldTabStr)
			oldTabStr = dosTabStr;
	}
	SysWaitStart();	
	oldTabArr = tabAToI(oldTabStr);
	newTabArr = tabAToI(newTabStr);
	MarkSavePos();
	MovEOF();
	totalLines = BufQCurrentLine();
	MovToOffset(0);
	do
	{
		MovHome();
		if (oldBuf = BufReadStr(0x7FFF))
		{
			if ((newBuf	= ReTabLine(oldBuf, oldTabArr, newTabArr, useSpaces)) 
				!= oldBuf)
			{
				BufDelToEOL();
				BufInsertStr(newBuf);
				StrFree(newBuf);
			}
			StrFree(oldBuf);
			if (!(curLine %	64))
			{
				MsgPrintf( MESSAGE_MSG_LEVEL, "Retabbing (line %ld of %ld)...",
					curLine, totalLines	);
			}
		}
		++curLine;
	} while	(MovDown(1)	== 1);
	MemFree(oldTabArr);
	MemFree(newTabArr);
	MarkRestorePos();
	MsgMessage(	"Retabbing complete." );
	SysWaitEnd();	
}

void DLL
EnTabBuffer(LPSTR tabStr)
{
	LPSTR newBuf;   
	LPSTR oldBuf;   
	long  curLine    = 1;
	long  totalLines = 0;
	LPINT tabArr;   
	
	if (ArgCount() == 0 || !tabStr)
		tabStr = BufQTabStr();
	SysWaitStart();	
	tabArr = tabAToI(tabStr);
	MarkSavePos();
	MovEOF();
	totalLines = BufQCurrentLine();
	MovToOffset(0);
	do
	{
		MovHome();
		if (oldBuf = BufReadStr(0x7FFF))
		{
			if ((newBuf	= EnTabLine(oldBuf,	tabArr)) !=	oldBuf)
			{
				BufDelToEOL();
				BufInsertStr(newBuf);
				StrFree(newBuf);
			}
			StrFree(oldBuf);
			if (!(curLine %	64))
			{
				MsgPrintf( MESSAGE_MSG_LEVEL, "Entabbing (line %ld of %ld)...",
					curLine, totalLines	);
			}
		}
		++curLine;
	} while	(MovDown(1)	== 1);
	MemFree(tabArr);
	MarkRestorePos();
	MsgMessage(	"Entabbing complete." );
	SysWaitEnd();	
}

void DLL
DeTabBuffer(LPSTR tabStr)
{
	LPSTR newBuf;
	LPSTR oldBuf;
	long curLine = 1;
	long totalLines	= 0;
	LPINT tabArr;
	
	if (ArgCount() == 0 || !tabStr)
		tabStr = BufQTabStr();
	SysWaitStart();	
	tabArr = tabAToI(tabStr);
	MarkSavePos();
	MovEOF();
	totalLines = BufQCurrentLine();
	MovToOffset(0);
	do
	{
		MovHome();
		oldBuf = BufReadStr(0x7FFF);
		if ((newBuf	= DeTabLine(oldBuf,	tabArr)) !=	oldBuf)
		{
			BufDelToEOL();
			BufInsertStr(newBuf);
			StrFree(newBuf);
		}
		StrFree(oldBuf);
		if (!(curLine %	64))
		{
			MsgPrintf( MESSAGE_MSG_LEVEL, "Detabbing (line %ld of %ld)...",
				curLine, totalLines);
		}
		++curLine;
	} while	(MovDown(1)	== 1);
	MemFree(tabArr);
	MarkRestorePos();
	MsgMessage(	"Detabbing complete." );
	SysWaitEnd();	
}

static LPSTR
EnTabLine(LPSTR	iLine, LPINT tabArr)
{
	LPSTR oLine;     
	LPSTR iP;        
	LPSTR oP;        
	int   oldIndent   = 1;    // column on which text begins
	int   newIndent   = 0;    
	int   newNumTabs  = 0;    // column on which text would begin if it were indented with tabs only
	int   n           = 1;    
	int   indentDelta = 0;    
	int   tabIdx      = 0;    
	BOOL  hadSpaces   = FALSE;
	
	if (!iLine || !*iLine)
		return(iLine);
	
	for	(iP	= iLine; *iP ==	' '	|| *iP == '\t';	++iP)
	{
		if (*iP	== ' ')
		{
			oldIndent++;
			hadSpaces =	TRUE;
		}
		else
		{
			oldIndent =	nextTab(oldIndent, tabArr);
			if (oldIndent <	0)
			{
				break;
			}
		}	
	}	
	if (!hadSpaces)
	{
		return(iLine);
	}
	newIndent =	0;
	for	(tabIdx	= 0; tabArr[tabIdx]; ++tabIdx)
	{
		if (tabArr[tabIdx] > oldIndent)
			break;
	}
	if (tabIdx &&  tabArr[tabIdx] > oldIndent)
	{
		newNumTabs = tabIdx;
		newIndent =	tabArr[tabIdx -	1];
	}
	if (!newIndent)
	{
		for (n = 0; ((n = nextTab(n, tabArr)) <= oldIndent) && n > 0; )
		{
			newNumTabs++;
			newIndent =	n;
		}
	}
	oLine =	MemAlloc(strlen(iLine) + 1);
	oP = oLine;	
	while (newNumTabs--)
		*oP++ =	'\t';

	if (!newIndent)		// current indentation is less than	one	tabstop
	{
		newIndent =	1;	// first column	is number 1
	}
	indentDelta	= oldIndent	- newIndent;	
	if (indentDelta	> 0)
		memset(oP, ' ',	indentDelta);
	strcpy(oP +	indentDelta, iP);
	return(oLine);
}

static LPSTR
DeTabLine(LPSTR	iLine, LPINT tabArr)
{
	LPSTR oLine;	
	LPSTR iP;		
	LPSTR oP;		
	int	  oldIndent	 = 1;
	int	  newIndent	 = 1;
	int	  newNumTabs = 0;
	BOOL  hadTabs =	FALSE;
	
	if (!iLine || !*iLine)
		return(iLine);
	for	(iP	= iLine; *iP ==	' '	|| *iP == '\t';	++iP)
	{
		if (*iP	== ' ')
			oldIndent++;
		else
		{
			hadTabs	= TRUE;
			if ((oldIndent = nextTab(oldIndent,	tabArr)) < 0)
				break;
		}	
	}
	if (!hadTabs)
	{
		return(iLine);
	}
	oP = oLine = MemAlloc(oldIndent	+ strlen(iP) + 1);
	if (oldIndent >	1)
	{
		memset(oP, ' ',	oldIndent -	1);
		strcpy(oP +	oldIndent -	1, iP);
	}
	return(oLine);
}


static LPSTR
ReTabLine(LPSTR iLine, LPINT oldTabArr, LPINT newTabArr, BOOL useSpaces)
{
	LPSTR oLine;     
	LPSTR iP;        
	LPSTR oP;        
	int   oldIndent   = 1;    // column on which text begins
	int   newIndent   = 0;    
	int   newNumTabs  = 0;    // column on which text would begin if it were indented with tabs only
	int   n           = 1;    
	int   indentDelta = 0;    
	int   tabIdx      = 0;    
	BOOL  hadTabs     = FALSE;
	
	if (!iLine || !*iLine)
		return(iLine);
		
	for	(iP	= iLine; *iP ==	' '	|| *iP == '\t';	++iP)
	{
		if (*iP	== ' ')
			oldIndent++;
		else
		{
			hadTabs	= TRUE;
			if ((oldIndent = nextTab(oldIndent,	oldTabArr)) < 0)
				break;
		}	
	}
	if (!hadTabs && !useSpaces)
	{
		return(iLine);
	}
	oP = oLine = MemAlloc(oldIndent	+ strlen(iP) + 1);

	newIndent =	0;

	for	(tabIdx	= 0; newTabArr[tabIdx]; ++tabIdx)
	{
		if (newTabArr[tabIdx] > oldIndent)
			break;
	}
	if (tabIdx && newTabArr[tabIdx] > oldIndent)
	{
		newNumTabs = tabIdx;
		newIndent =	newTabArr[tabIdx - 1];
	}
	if (!newIndent)
	{
		for (n = 0; ((n = nextTab(n, newTabArr)) <= oldIndent) && n > 0; )
		{
			newNumTabs++;
			newIndent =	n;
		}
	}

	if (!useSpaces)
	{
		while (newNumTabs--)
			*oP++ =	'\t';

		if (newIndent == oldIndent)
		{
			MemFree(oLine);
			return(iLine);
		}
		if (!newIndent)		// current indentation is less than	one	tabstop
		{
			newIndent =	1;	// first column	is number 1
		}
		indentDelta	= oldIndent	- newIndent;	
		if (indentDelta	> 0)
			memset(oP, ' ',	indentDelta);
		strcpy(oP +	indentDelta, iP);
	}
	else
	{
		if (!newIndent)
			newIndent = 1;
		--newIndent;	
		memset(oP, ' ',	newIndent);
		strcpy(oP +	newIndent, iP);
		
	}

	return(oLine);
}


static LPINT
tabAToI(LPSTR tabStr)
{
	LPINT tabArr;
	LPSTR tabSP;
	LPINT tabAP;
	
	/*
	 *	Use	strlen(tabStr) as an upper bound for the number
	 *	of tab settings.
	 */
	if (!(tabArr = MemCalloc(strlen(tabStr), sizeof(int))))
		return(NULL);
		
	tabSP =	tabStr;
	tabAP =	tabArr;
	
	while (*tabSP)
	{
		while (isspace(*tabSP))
			++tabSP;
		if (!*tabSP)
			break;
		*tabAP++ = atoi(tabSP);
		while (*tabSP && !isspace(*tabSP))
			++tabSP;
	}
	*tabAP = 0;
	return(tabArr);
}

int DLL
NextTab(int oldCol, LPSTR tabStr)
{
	LPINT tabArr;
	int retval = -1;
	int argcount = ArgCount();
	
	if (argcount >= 0)
	{
		if (argcount < 2)
			tabStr = BufQTabStr();
		if (argcount < 1)
			oldCol = (int)BufQCurrentCol();
	}
		
	if (tabArr = tabAToI(tabStr))
	{
		retval = nextTab(oldCol, tabArr);
		MemFree(tabArr);
	}
	return(retval);
}

static int
nextTab(int	oldCol,	LPINT tabArr)
{
	int	  newCol;  			   
	int	  lastStop;
	LPINT tabp;	   
	char  interval;

	if (!tabArr)	// bad tabArr argument
		return(-1);

	for	(tabp =	tabArr;	*tabp; ++tabp)	// is there	an explicit	stop?
	{
		if (*tabp >	oldCol)
			break;
	}
	
	// If there	was	an explicit	stop, return it.
	if (*tabp)
		return(*tabp);
	
	// If not, calculate from interval
	if (--tabp > tabArr	&& --tabp >= tabArr)	// find	interval
	{
		interval = tabp[1] - tabp[0];
		lastStop = tabp[1];
	}
	else
	{
		interval = tabArr[0];
		lastStop = tabArr[0];
	}
	if (lastStop > oldCol)
	{
		return(lastStop);
	}
	
	if (interval <=	0)
	{
		return(-2);
	}
/*
 * Add intervals to the last explicit stop until we find
 * a stop beyond the current column.
 * 
 * newCol = laststop + interval * I
 * I = (oldCol + interval) / interval
 */
	newCol = lastStop + (interval * (((oldCol - lastStop) + interval) / interval));
	return(newCol);
}

BOOL DLL
TabInstallHandler(void)
{
	if (!tabHashPtr)
		if (!(tabHashPtr = HashCreateTable(41, sizeof(int), HASH_IGN_CASE)))
			return(FALSE);
	if (!eva[0])
		eva[0] = EventRegister (EVENT_NEW_EDITFILE, EVENT_NORMAL, "Tab_new_handler");
	if (!eva[1])
		eva[1] = EventRegister (EVENT_SAVE_BUFFER, EVENT_NORMAL, "Tab_save_handler");
	if (!eva[2])
		eva[2] = EventRegister (EVENT_BUFFER_SAVE_REOPENED, EVENT_NORMAL, "Tab_saved_handler");
	if (eva[0] && eva[1] && eva[2])
	{
		return(TRUE);
	}
	return(FALSE);
} 

BOOL DLL
TabRemoveHandler(void)
{
	if (tabHashPtr)
	{
		HashDeleteTable(tabHashPtr, NULL);
		tabHashPtr = NULL;
	}
	
	if (eva[0])
	{
		EventDeregister(eva[0]);
		eva[0] = 0;
	}
	if (eva[1])
	{
		EventDeregister(eva[1]);
		eva[1] = 0;
	}
	if (eva[2])
	{
		EventDeregister(eva[2]);
		eva[2] = 0;
	}
	if (!eva[0] && !eva[1] && !eva[2])
		return(TRUE);
	return(FALSE);	
} 

/* Public functions:
 *
 * The NEW_EDITFILE event handler is called whenever a file is opened for
 * editing.  It retabifies from the DOS 8-column tabs to whatever the 
 * buffer settings demand.
 *
 * The BUFFER_SAVED handler is called whenever a buffer is written to disk.
 * It retabifies to 8-column tabs.
 *
 * The purpose of this handler is to allow the use of Codewright to
 * edit files containing standard (8-column) tabs transparently.
 */ 

long DLL
Tab_new_handler (long event, long data)
{
	LPINT hp;      
	LPSTR fileName;
	BOOL  reTab     = FALSE;

	if (!tabHashPtr)
		return(0);
	if (BufQIsSystem(BufQCurrentBuffer()))
		return(0); 

	if (tabConfirm)
		reTab = MsgNotifyConfirm("Retabify this document according to your buffer properties?");
	else
		reTab = TRUE;	
	if (fileName = BufQFilename()) 
	{ 
		if (hp = (int *)HashGetEntry(tabHashPtr, fileName, strlen(fileName)))
		{
			if (reTab == TRUE)
			{
				BOOL spacesOnly;
				
				if (tabSpacesOnly > 0)
					spacesOnly = TRUE;
				else if (tabSpacesOnly < 0)
					spacesOnly = FALSE;
				else
					spacesOnly = !BufQTabUsage();
				ReTabBuffer(dosTabStr, BufQTabStr(), spacesOnly);	
				*hp = RETAB;
			}
			else
			{
				*hp = NORETAB;
			}
		}
	}
	return(0);
}
 
long DLL 
Tab_save_handler(long event, long data) 
{ 
	LPSTR fileName; 
	LPINT hp; 
	BOOL reTab = FALSE;
	 
	if (!tabHashPtr) 
		return(0); 
	if (BufQIsSystem(BufQCurrentBuffer()))
		return(0); 

	if (fileName = BufQFilename())
		if (tabHashPtr)
			hp = (LPINT)HashFindEntry(tabHashPtr, fileName, strlen(fileName));
 
	if (hp && *hp == RETAB)
	{ 
		if (tabConfirm)
			reTab = MsgNotifyConfirm("Retabify this document for DOS?");
		else
			reTab = TRUE;	
		if (!reTab)
   			return(0);
		
		ReTabBuffer(BufQTabStr(), dosTabStr, FALSE);	
	} 
	 
	return(0); 
}
 
 
long DLL 
Tab_saved_handler(long event, long data) 
{ 
	LPSTR fileName; 
	LPINT hp; 
	 
	if (!tabHashPtr) 
		return(0); 
	if (BufQIsSystem(BufQCurrentBuffer()))
		return(0); 
	 
	if (fileName = BufQFilename())
		if (tabHashPtr)
			hp = (LPINT)HashFindEntry(tabHashPtr, fileName, strlen(fileName));
 
	if (hp && *hp == RETAB)
	{ 
		ReTabBuffer(dosTabStr, BufQTabStr(), FALSE);	
	} 
	 
	return(0); 
}

/*
 * Tells this package whether to convert the current buffer's 
 * tabs or not.
 */
 
BOOL DLL
TabSetConvert(BOOL convert)
{
	LPSTR fileName; 
	LPINT hp;
	BOOL retval = TRUE;
	
	if (!tabHashPtr) 
		return(0); 
	if (BufQIsSystem(BufQCurrentBuffer()))
		return(FALSE);
	convert = !!convert;	
	if (ArgCount() == 1)
		convert = TRUE;
	if (fileName = BufQFilename())
	{ 
		/* see if we know about this buffer already	*/
		if (hp = (LPINT)HashFindEntry(tabHashPtr, fileName, strlen(fileName)))
			*hp = convert ? RETAB : NORETAB;
		/* we don't.  Create a new entry for it.	*/	
		else if (hp = (int *)HashGetEntry(tabHashPtr, fileName, strlen(fileName))) 
			*hp = convert ? RETAB : NORETAB;
		else
			retval = FALSE;	
	}
	else
		retval = FALSE;
	return(retval);	
}

int
tabLookup(void) 
{ 
	LPSTR fileName; 
	LPINT hp; 
	 
	if (fileName = BufQFilename())
		if (tabHashPtr)
			if (hp = (LPINT)HashFindEntry(tabHashPtr, fileName, strlen(fileName)))
				return(*hp); 
	return(-1); 
} 
 
